@@ -1,5 +1,5 @@
 /*
- * $Id: ngrep.c,v 1.66 2005/02/20 05:52:42 jpr5 Exp $
+ * $Id: ngrep.c,v 1.78 2005/06/30 23:35:20 jpr5 Exp $
  *
  * Copyright (c) 2005  Jordan Ritter <jpr5@darkridge.com>
  *
@@ -54,13 +54,19 @@
 #include <io.h>
 #include <getopt.h>
 #include <winsock2.h>
-#include <nettypes.h>
+#include <types.h>
+#include <inet_ntop.h>
+
+#define strcasecmp stricmp
+
 #else
+
 #include <netinet/ip.h>
 #include <netinet/tcp.h>
 #include <netinet/udp.h>
 #include <netinet/ip_icmp.h>
 #include <netinet/igmp.h>
+
 #endif
 
 #include <stdlib.h>
@@ -78,6 +84,11 @@
 #include "config.h"
 #endif
 
+#if USE_IPv6 && !defined(_WIN32)
+#include <netinet/ip6.h>
+#include <netinet/icmp6.h>
+#endif
+
 #if USE_PCRE
 #include "pcre-5.0/pcre.h"
 #else
@@ -87,64 +98,89 @@
 #include "ngrep.h"
 
 
+static char rcsver[] = "$Revision: 1.78 $";
+
+/*
+ * Configuration Options
+ */
+
+uint16_t snaplen = 65535, limitlen = 65535, promisc = 1, to = 100;
+uint16_t match_after = 0, keep_matching = 0, matches = 0, max_matches = 0;
+
+uint8_t  re_match_word = 0, re_ignore_case = 0, re_multiline_match = 1;
+uint8_t  show_empty = 0, show_hex = 0, show_proto = 0, quiet = 0;
+uint8_t  invert_match = 0, bin_match = 0;
+uint8_t  live_read = 1, want_delay = 0;
+uint8_t  dont_dropprivs = 0;
+
+char *read_file = NULL, *dump_file = NULL;
+char *usedev = NULL;
+
+char nonprint_char = '.';
+
+/*
+ * GNU Regex/PCRE
+ */
+
+#if USE_PCRE
+int32_t err_offset;
+char *re_err = NULL;
+
+pcre *pattern = NULL;
+pcre_extra *pattern_extra = NULL;
+#else
+const char *re_err = NULL;
+
+struct re_pattern_buffer pattern;
+#endif
+
+/*
+ * Matching
+ */
+
+char *match_data = NULL, *bin_data = NULL;
+uint16_t match_len = 0;
+int8_t (*match_func)() = &blank_match_func;
+
+int8_t dump_single = 0;
+void (*dump_func)(unsigned char *, uint32_t) = &dump_formatted;
+
+/*
+ * BPF/Network
+ */
+
+char *filter = NULL, *filter_file = NULL;
+char pc_err[PCAP_ERRBUF_SIZE];
+uint8_t link_offset;
+
+pcap_t *pd = NULL;
+pcap_dumper_t *pd_dump = NULL;
+struct bpf_program pcapfilter;
+struct in_addr net, mask;
+
+/*
+ * Timestamp/delay functionality
+ */
+
+struct timeval prev_ts = {0, 0}, prev_delay_ts = {0,0};
 #if defined(_WIN32)
-#define strcasecmp stricmp
 struct timeval delay_tv;
 FD_SET delay_fds;
 SOCKET delay_socket = 0;
 #endif
 
-static char rcsver[] = "$Revision: 1.66 $";
-
-unsigned snaplen = 65535, limitlen = 65535, promisc = 1, to = 1000;
-unsigned show_empty = 0, show_hex = 0, quiet = 0;
-unsigned match_after = 0, keep_matching = 0;
-unsigned invert_match = 0, bin_match = 0;
-unsigned matches = 0, max_matches = 0;
-unsigned live_read = 1, want_delay = 0;
-unsigned dont_dropprivs = 0;
-
-char nonprint_char = '.';
-
-char pc_err[PCAP_ERRBUF_SIZE];
-#if USE_PCRE
-int err_offset;
-char *re_err = NULL;
-#else
-const char *re_err = NULL;
-#endif
-
-unsigned re_match_word = 0, re_ignore_case = 0, re_multiline_match = 1;
-
-#if USE_PCRE
-pcre *pattern = NULL;
-pcre_extra *pattern_extra = NULL;
-#else
-struct re_pattern_buffer pattern;
-#endif
-
-char *match_data = NULL, *bin_data = NULL, *filter = NULL, *filter_file = NULL;
-int (*match_func)() = &blank_match_func;
-void (*dump_func)(char *, unsigned) = &dump_formatted;
-unsigned match_len = 0;
-
-struct bpf_program pcapfilter;
-struct in_addr net, mask;
-pcap_t *pd = NULL;
-char *usedev = NULL;
-unsigned link_offset;
-
-char *read_file = NULL, *dump_file = NULL;
-pcap_dumper_t *pd_dump = NULL;
-
-struct timeval prev_ts = {0, 0}, prev_delay_ts = {0,0};
 void (*print_time)() = NULL, (*dump_delay)() = dump_delay_proc_init;
 
-unsigned ws_row, ws_col = 80, ws_col_forced = 0;
+
+/*
+ * When !Win32, windowsize stuff
+ */
+
+uint32_t ws_row, ws_col = 80, ws_col_forced = 0;
 
 
 int main(int argc, char **argv) {
-    int c;
+    int32_t c;
 
     signal(SIGINT,   clean_exit);
     signal(SIGABRT,  clean_exit);
@@ -155,7 +191,7 @@ int main(int argc, char **argv) {
     signal(SIGWINCH, update_windowsize);
 #endif
 
-    while ((c = getopt(argc, argv, "LhXViwqpevxlDtTRMs:n:c:d:A:I:O:S:P:F:W:")) != EOF) {
+    while ((c = getopt(argc, argv, "LNhXViwqpevxlDtTRMs:n:c:d:A:I:O:S:P:F:W:")) != EOF) {
         switch (c) {
             case 'W': {
                 if (!strcasecmp(optarg, "normal"))
@@ -164,7 +200,10 @@ int main(int argc, char **argv) {
                     dump_func = &dump_byline;
                 else if (!strcasecmp(optarg, "none"))
                     dump_func = &dump_unwrapped;
-                else {
+                else if (!strcasecmp(optarg, "single")) {
+                    dump_func = &dump_unwrapped;
+                    dump_single = 1;
+                } else {
                     printf("fatal: unknown wrap method '%s'\n", optarg);
                     usage(-1);
                 }
@@ -221,7 +260,7 @@ int main(int argc, char **argv) {
             case 'T':
                 print_time = &print_time_diff;
 #if defined(_WIN32)
-                prev_ts.tv_sec  = (long)time(NULL);
+                prev_ts.tv_sec  = (uint32_t)time(NULL);
                 prev_ts.tv_usec = 0;
 #else
                 gettimeofday(&prev_ts, NULL);
@@ -262,6 +301,9 @@ int main(int argc, char **argv) {
             case 'X':
                 bin_match++;
                 break;
+            case 'N':
+                show_proto++;
+                break;
             case 'h':
                 usage(0);
             default:
@@ -307,7 +349,7 @@ int main(int argc, char **argv) {
             memset(&mask, 0, sizeof(mask));
         }
 
-        if (!quiet) {
+        if (quiet < 2) {
             printf("interface: %s", dev);
             if (net.s_addr && mask.s_addr) {
                 printf(" (%s/", inet_ntoa(net));
@@ -356,7 +398,7 @@ int main(int argc, char **argv) {
         }
 
     } else {
-        char *default_filter = "ip";
+        char *default_filter = BPF_FILTER_IP;
 
         if (pcap_compile(pd, &pcapfilter, default_filter, 0, mask.s_addr)) {
             pcap_perror(pd, "pcap compile");
@@ -364,7 +406,7 @@ int main(int argc, char **argv) {
         }
     }
 
-    if (filter && !quiet)
+    if (filter && quiet < 2)
         printf("filter: %s\n", filter);
 
     if (pcap_setfilter(pd, &pcapfilter)) {
@@ -374,16 +416,16 @@ int main(int argc, char **argv) {
 
     if (match_data) {
         if (bin_match) {
-            unsigned i = 0, n;
+            uint32_t i = 0, n;
+            uint32_t len;
             char *s, *d;
-            unsigned len;
 
             if (re_match_word || re_ignore_case) {
                 fprintf(stderr, "fatal: regex switches are incompatible with binary matching\n");
                 clean_exit(-1);
             }
 
-            len = (unsigned)strlen(match_data);
+            len = (uint32_t)strlen(match_data);
             if (len % 2 != 0 || !strishex(match_data)) {
                 fprintf(stderr, "fatal: invalid hex string specified\n");
                 clean_exit(-1);
@@ -394,7 +436,7 @@ int main(int argc, char **argv) {
             d = bin_data;
 
             if ((s = strchr(match_data, 'x')))
-                len -= (unsigned)(++s - match_data - 1);
+                len -= (uint32_t)(++s - match_data - 1);
             else s = match_data;
 
             while (i <= len) {
@@ -409,7 +451,7 @@ int main(int argc, char **argv) {
         } else {
 
 #if USE_PCRE
-            unsigned pcre_options = PCRE_UNGREEDY;
+            uint32_t pcre_options = PCRE_UNGREEDY;
 
             if (re_ignore_case)
                 pcre_options |= PCRE_CASELESS;
@@ -424,8 +466,8 @@ int main(int argc, char **argv) {
                 re_syntax_options |= RE_DOT_NEWLINE;
 
             if (re_ignore_case) {
+                uint32_t i;
                 char *s;
-                unsigned i;
 
                 pattern.translate = (char*)malloc(256);
                 s = pattern.translate;
@@ -474,7 +516,7 @@ int main(int argc, char **argv) {
             match_func = &re_match_func;
         }
 
-        if (!quiet && match_data && strlen(match_data))
+        if (quiet < 2 && match_data && strlen(match_data))
             printf("%smatch: %s%s\n", invert_match?"don't ":"",
                    (bin_data && !strchr(match_data, 'x'))?"0x":"", match_data);
     }
@@ -551,264 +593,207 @@ int main(int argc, char **argv) {
     while (pcap_loop(pd, 0, (pcap_handler)process, 0));
 
     clean_exit(0);
+
+    /* NOT REACHED */
+    return 0;
 }
 
-
 void process(u_char *d, struct pcap_pkthdr *h, u_char *p) {
-    struct ip *ip_packet = (struct ip *)(p + link_offset);
+    struct ip      *ip_packet  = (struct ip *)    (p + link_offset);
+#if USE_IPv6
+    struct ip6_hdr *ip6_packet = (struct ip6_hdr*)(p + link_offset);
+#endif
+
+    uint32_t ip_ver   = ip_packet->ip_v;
+
+    uint8_t  ip_proto = 0;
+    uint32_t ip_hl    = 0;
+    uint32_t ip_off   = 0;
+
+    uint8_t  fragmented  = 0;
+    uint16_t frag_offset = 0;
+    uint32_t frag_id     = 0;
+
+    char ip_src[INET6_ADDRSTRLEN + 1],
+         ip_dst[INET6_ADDRSTRLEN + 1];
+
+    unsigned char *data;
+    uint32_t len = 0;
+
+    switch (ip_ver) {
+
+        case 4: {
 
 #if defined(AIX)
 #undef ip_hl
-    unsigned ip_hl = ip_packet->ip_ff.ip_fhl*4;
+            ip_hl       = ip_packet->ip_ff.ip_fhl * 4;
 #else
-    unsigned ip_hl = ip_packet->ip_hl*4;
+            ip_hl       = ip_packet->ip_hl * 4;
 #endif
+            ip_proto    = ip_packet->ip_p;
+            ip_off      = ntohs(ip_packet->ip_off);
 
-    unsigned ip_off      = ntohs(ip_packet->ip_off);
-    unsigned fragmented  = ip_off & (IP_MF | IP_OFFMASK);
-    unsigned frag_offset = fragmented?(ip_off & IP_OFFMASK) * 8:0;
+            fragmented  = ip_off & (IP_MF | IP_OFFMASK);
+            frag_offset = (fragmented) ? (ip_off & IP_OFFMASK) * 8 : 0;
+            frag_id     = ntohs(ip_packet->ip_id);
 
-    char *data;
-    unsigned len;
+            inet_ntop(AF_INET, (const void *)&ip_packet->ip_src, ip_src, sizeof(ip_src));
+            inet_ntop(AF_INET, (const void *)&ip_packet->ip_dst, ip_dst, sizeof(ip_dst));
+        } break;
 
-    switch (ip_packet->ip_p) {
+#if USE_IPv6
+        case 6: {
+
+            ip_hl    = sizeof(struct ip6_hdr);
+            ip_proto = ip6_packet->ip6_nxt;
+
+            if (ip_proto == IPPROTO_FRAGMENT) {
+                struct ip6_frag *ip6_fraghdr;
+
+                ip6_fraghdr = (struct ip6_frag *)(((unsigned char *)ip6_packet) + ip_hl);
+                ip_hl      += sizeof(struct ip6_frag);
+                ip_proto    = ip6_fraghdr->ip6f_nxt;
+
+                fragmented  = 1;
+                frag_offset = ntohs(ip6_fraghdr->ip6f_offlg & IP6F_OFF_MASK);
+                frag_id     = ntohl(ip6_fraghdr->ip6f_ident);
+            }
+
+            inet_ntop(AF_INET6, (const void *)&ip6_packet->ip6_src, ip_src, sizeof(ip_src));
+            inet_ntop(AF_INET6, (const void *)&ip6_packet->ip6_dst, ip_dst, sizeof(ip_dst));
+        } break;
+#endif
+    }
+
+    if (quiet < 1) {
+        printf("#");
+        fflush(stdout);
+    }
+
+    switch (ip_proto) {
         case IPPROTO_TCP: {
-            struct tcphdr *tcp = (struct tcphdr *)(((char *)ip_packet) + ip_hl);
-            unsigned tcphdr_offset = fragmented?0:(tcp->th_off * 4);
+            struct tcphdr *tcp     = (struct tcphdr *)(((unsigned char *)ip_packet) + ip_hl);
+            uint16_t tcphdr_offset = (frag_offset) ? 0 : (tcp->th_off * 4);
 
-            if (!quiet) {
-                printf("#");
-                fflush(stdout);
+            data = ((unsigned char *)tcp) + tcphdr_offset;
+
+            switch (ip_ver) {
+                case 4: {
+                    if ((len = ntohs(ip_packet->ip_len)) < h->caplen)
+                        len -= ip_hl + tcphdr_offset;
+                    else
+                        len = h->caplen - link_offset - ip_hl - tcphdr_offset;
+                } break;
+
+#if USE_IPv6
+                case 6: {
+                    if ((len = ntohs(ip6_packet->ip6_plen + ip_hl)) < h->caplen)
+                        len -= ip_hl + tcphdr_offset;
+                    else
+                        len = h->caplen - link_offset - ip_hl - tcphdr_offset;
+                } break;
+#endif
             }
 
-            data = ((char*)tcp) + tcphdr_offset;
-
-            if ((len = ntohs(ip_packet->ip_len)) < h->caplen)
-                len -= ip_hl + tcphdr_offset;
-            else len = h->caplen - link_offset - ip_hl - tcphdr_offset;
-
-            if (len > limitlen) len = limitlen;
-
-            if (((len || show_empty) && (((int)(*match_func)(data, len)) != invert_match))
-                || keep_matching) {
-
-                if (!live_read && want_delay)
-                    dump_delay(h);
-
-                printf("\nT ");
-
-                if (print_time)
-                    print_time(h);
-
-                if (tcphdr_offset || !frag_offset) {
-                    printf("%s:%u -", inet_ntoa(ip_packet->ip_src), ntohs(tcp->th_sport));
-                    printf("> %s:%u", inet_ntoa(ip_packet->ip_dst), ntohs(tcp->th_dport));
-                    printf(" [%s%s%s%s%s%s%s%s]",
-                           (tcp->th_flags & TH_ACK)?"A":"",
-                           (tcp->th_flags & TH_SYN)?"S":"",
-                           (tcp->th_flags & TH_RST)?"R":"",
-                           (tcp->th_flags & TH_FIN)?"F":"",
-                           (tcp->th_flags & TH_URG)?"U":"",
-                           (tcp->th_flags & TH_PUSH)?"P":"",
-                           (tcp->th_flags & TH_ECE)?"E":"",
-                           (tcp->th_flags & TH_CWR)?"C":"");
-                } else {
-                    printf("%s -", inet_ntoa(ip_packet->ip_src));
-                    printf("> %s", inet_ntoa(ip_packet->ip_dst));
-                }
-
-                if (fragmented)
-                    printf(" %s%u@%u:%u\n", frag_offset?"+":"", ntohs(ip_packet->ip_id),
-                           frag_offset, len);
-                else printf("\n");
-
-                if (pd_dump)
-                    pcap_dump((u_char*)pd_dump, h, p);
-
-                if (quiet < 2)
-                    dump_func(data, len);
-            }
+            dump_packet(h, p, ip_proto, data, len,
+                        ip_src, ip_dst, ntohs(tcp->th_sport), ntohs(tcp->th_dport), tcp->th_flags,
+                        tcphdr_offset, fragmented, frag_offset, frag_id);
         } break;
 
         case IPPROTO_UDP: {
-            struct udphdr* udp = (struct udphdr *)(((char *)ip_packet) + ip_hl);
-            unsigned udphdr_offset = (fragmented)?0:sizeof(struct udphdr);
+            struct udphdr *udp     = (struct udphdr *)(((unsigned char *)ip_packet) + ip_hl);
+            uint16_t udphdr_offset = (frag_offset) ? 0 : sizeof(struct udphdr);
 
-            if (!quiet) {
-                printf("#");
-                fflush(stdout);
-            }
+            data = ((unsigned char *)udp) + udphdr_offset;
 
-            data = ((char*)udp) + udphdr_offset;
+            switch (ip_ver) {
+                case 4: {
+                    if ((len = ntohs(ip_packet->ip_len)) < h->caplen)
+                        len -= ip_hl + udphdr_offset;
+                    else
+                        len = h->caplen - link_offset - ip_hl - udphdr_offset;
+                } break;
 
-            if ((len = ntohs(ip_packet->ip_len)) < h->caplen)
-                len -= ip_hl + udphdr_offset;
-            else len = h->caplen - link_offset - ip_hl - udphdr_offset;
-
-            if (len > limitlen) len = limitlen;
-
-            if (((len || show_empty) && (((int)(*match_func)(data, len)) != invert_match))
-                || keep_matching) {
-
-                if (!live_read && want_delay)
-                    dump_delay(h);
-
-                printf("\nU ");
-
-                if (print_time)
-                    print_time(h);
-
-                if (udphdr_offset || !frag_offset) {
-#if HAVE_DUMB_UDPHDR
-                    printf("%s:%u -", inet_ntoa(ip_packet->ip_src), ntohs(udp->source));
-                    printf("> %s:%u", inet_ntoa(ip_packet->ip_dst), ntohs(udp->dest));
-#else
-                    printf("%s:%u -", inet_ntoa(ip_packet->ip_src), ntohs(udp->uh_sport));
-                    printf("> %s:%u", inet_ntoa(ip_packet->ip_dst), ntohs(udp->uh_dport));
+#if USE_IPv6
+                case 6: {
+                    if ((len = ntohs(ip6_packet->ip6_plen + ip_hl)) < h->caplen)
+                        len -= ip_hl + udphdr_offset;
+                    else
+                        len = h->caplen - link_offset - ip_hl - udphdr_offset;
+                } break;
 #endif
-                } else {
-                    printf("%s -", inet_ntoa(ip_packet->ip_src));
-                    printf("> %s", inet_ntoa(ip_packet->ip_dst));
-                }
-
-                if (fragmented)
-                    printf(" %s%u@%u:%u\n", frag_offset?"+":"", ntohs(ip_packet->ip_id),
-                           frag_offset, len);
-                else printf("\n");
-
-                if (pd_dump)
-                    pcap_dump((u_char*)pd_dump, h, p);
-
-                if (quiet < 2)
-                    dump_func(data, len);
             }
+
+            dump_packet(h, p, ip_proto, data, len, ip_src, ip_dst,
+#if HAVE_DUMB_UDPHDR
+                        ntohs(udp->source), ntohs(udp->dest), 0,
+#else
+                        ntohs(udp->uh_sport), ntohs(udp->uh_dport), 0,
+#endif
+                        udphdr_offset, fragmented, frag_offset, frag_id);
         } break;
 
         case IPPROTO_ICMP: {
-            struct icmp* ic = (struct icmp *)(((char *)ip_packet) + ip_hl);
-            unsigned icmphdr_offset = fragmented?0:4;
+            struct icmp *ic4        = (struct icmp *)(((unsigned char *)ip_packet) + ip_hl);
+            uint16_t icmphdr_offset = (frag_offset) ? 0 : 4;
 
-            if (!quiet) {
-                printf("#");
-                fflush(stdout);
-            }
-
-            data = ((char*)ic) + icmphdr_offset;
+            data = ((unsigned char *)ic4) + icmphdr_offset;
 
             if ((len = ntohs(ip_packet->ip_len)) < h->caplen)
                 len -= ip_hl + icmphdr_offset;
-            else len = h->caplen - link_offset - ip_hl - icmphdr_offset;
+            else
+                len = h->caplen - link_offset - ip_hl - icmphdr_offset;
 
-            if (len > limitlen) len = limitlen;
-
-            if (((len || show_empty) && (((int)(*match_func)(data, len)) != invert_match))
-                || keep_matching) {
-
-                if (!live_read && want_delay)
-                    dump_delay(h);
-
-                printf("\nI ");
-
-                if (print_time)
-                    print_time(h);
-
-                printf("%s -", inet_ntoa(ip_packet->ip_src));
-                printf("> %s", inet_ntoa(ip_packet->ip_dst));
-
-                if (icmphdr_offset || !frag_offset)
-                    printf(" %u:%u", ic->icmp_type, ic->icmp_code);
-
-                if (fragmented)
-                    printf(" %s%u@%u:%u\n", frag_offset?"+":"", ntohs(ip_packet->ip_id),
-                           frag_offset, len);
-                else printf("\n");
-
-                if (pd_dump)
-                    pcap_dump((u_char*)pd_dump, h, p);
-
-                if (quiet < 2)
-                    dump_func(data, len);
-            }
+            dump_packet(h, p, ip_proto, data, len,
+                        ip_src, ip_dst, ic4->icmp_type, ic4->icmp_code, 0,
+                        icmphdr_offset, fragmented, frag_offset, frag_id);
         } break;
 
+#if USE_IPv6
+        case IPPROTO_ICMPV6: {
+            struct icmp6_hdr *ic6   = (struct icmp6_hdr *)(((unsigned char *)ip6_packet) + ip_hl);
+            uint16_t icmphdr_offset = (frag_offset) ? 0 : 4;
+
+            data = ((unsigned char *)ic6) + icmphdr_offset;
+
+            if ((len = ntohs(ip6_packet->ip6_plen + ip_hl)) < h->caplen)
+                len -= ip_hl + icmphdr_offset;
+            else
+                len = h->caplen - link_offset - ip_hl - icmphdr_offset;
+
+            dump_packet(h, p, ip_proto, data, len,
+                        ip_src, ip_dst, ic6->icmp6_type, ic6->icmp6_code, 0,
+                        icmphdr_offset, fragmented, frag_offset, frag_id);
+        } break;
+#endif
+
         case IPPROTO_IGMP: {
-            struct igmp* ig = (struct igmp *)(((char *)ip_packet) + ip_hl);
-            unsigned igmphdr_offset = fragmented?0:4;
+            struct igmp *ig         = (struct igmp *)(((unsigned char *)ip_packet) + ip_hl);
+            uint16_t igmphdr_offset = (frag_offset) ? 0 : 4;
 
-            if (!quiet) {
-                printf("#");
-                fflush(stdout);
-            }
-
-            data = ((char*)ig) + igmphdr_offset;
+            data = ((unsigned char *)ig) + igmphdr_offset;
 
             if ((len = ntohs(ip_packet->ip_len)) < h->caplen)
                 len -= ip_hl + igmphdr_offset;
-            else len = h->caplen - link_offset - ip_hl - igmphdr_offset;
+            else
+                len = h->caplen - link_offset - ip_hl - igmphdr_offset;
 
-            if (len > limitlen) len = limitlen;
-
-            if (((len || show_empty) && (((int)(*match_func)(data, len)) != invert_match))
-                || keep_matching) {
-
-                if (!live_read && want_delay)
-                    dump_delay(h);
-
-                printf("\nG ");
-
-                if (print_time)
-                    print_time(h);
-
-                printf("%s -", inet_ntoa(ip_packet->ip_src));
-                printf("> %s", inet_ntoa(ip_packet->ip_dst));
-
-                if (igmphdr_offset || !frag_offset)
-                    printf(" %u:%u", ig->igmp_type, ig->igmp_code);
-
-                if (fragmented)
-                    printf(" %s%u@%u:%u\n", frag_offset?"+":"", ntohs(ip_packet->ip_id),
-                           frag_offset, len);
-                else printf("\n");
-
-                if (pd_dump)
-                    pcap_dump((u_char*)pd_dump, h, p);
-
-                if (quiet < 2)
-                    dump_func(data, len);
-            }
+            dump_packet(h, p, ip_proto, data, len,
+                        ip_src, ip_dst, ig->igmp_type, ig->igmp_code, 0,
+                        igmphdr_offset, fragmented, frag_offset, frag_id);
         } break;
 
         default: {
-            data = (char*)(((char*)ip_packet) + ip_hl);
+            data = (((unsigned char *)ip_packet) + ip_hl);
 
             if ((len = ntohs(ip_packet->ip_len)) < h->caplen)
                 len -= ip_hl;
-            else len = h->caplen - link_offset - ip_hl;
+            else
+                len = h->caplen - link_offset - ip_hl;
 
-            if (len > limitlen) len = limitlen;
-
-            if (((len || show_empty) && (((int)(*match_func)(data, len)) != invert_match))
-                || keep_matching) {
-
-                if (!live_read && want_delay)
-                    dump_delay(h);
-
-                printf("\n? ");
-
-                if (print_time)
-                    print_time(h);
-
-                printf("%s -", inet_ntoa(ip_packet->ip_src));
-                printf("> %s ", inet_ntoa(ip_packet->ip_dst));
-
-                printf("[proto %u]\n", ip_packet->ip_p);
-
-                if (pd_dump)
-                    pcap_dump((u_char*)pd_dump, h, p);
-
-                if (quiet < 2)
-                    dump_func(data, len);
-            }
+            dump_packet(h, p, ip_proto, data, len,
+                        ip_src, ip_dst, 0, 0, 0,
+                        0, fragmented, frag_offset, frag_id);
         } break;
 
     }
@@ -820,10 +805,90 @@ void process(u_char *d, struct pcap_pkthdr *h, u_char *p) {
         keep_matching--;
 }
 
+void dump_packet(struct pcap_pkthdr *h, u_char *p, uint8_t proto, unsigned char *data, uint32_t len,
+                 const char *ip_src, const char *ip_dst, uint16_t sport, uint16_t dport, uint8_t flags,
+                 uint16_t hdr_offset, uint8_t frag, uint16_t frag_offset, uint32_t frag_id) {
 
-int re_match_func(char *data, unsigned len) {
+    if (!show_empty && len == 0)
+        return;
+
+    if (len > limitlen)
+        len = limitlen;
+
+    if ((len > 0 && match_func(data, len) == invert_match) && !keep_matching)
+        return;
+
+    if (!live_read && want_delay)
+        dump_delay(h);
+
+    {
+        char ident;
+
+        switch (proto) {
+            case IPPROTO_TCP:    ident = TCP;     break;
+            case IPPROTO_UDP:    ident = UDP;     break;
+            case IPPROTO_ICMP:   ident = ICMP;    break;
+            case IPPROTO_ICMPV6: ident = ICMPv6;  break;
+            case IPPROTO_IGMP:   ident = IGMP;    break;
+            default:             ident = UNKNOWN; break;
+        }
+
+        printf("\n%c", ident);
+    }
+
+    if (show_proto)
+        printf("(%u)", proto);
+
+    printf(" ");
+
+    if (print_time)
+        print_time(h);
+
+    if ((proto == IPPROTO_TCP || proto == IPPROTO_UDP) && (sport || dport) && (hdr_offset || frag_offset == 0))
+
+        printf("%s:%u -> %s:%u", ip_src, sport, ip_dst, dport);
+
+    else
+
+        printf("%s -> %s", ip_src, ip_dst);
+
+    if (proto == IPPROTO_TCP && flags)
+        printf(" [%s%s%s%s%s%s%s%s]",
+               (flags & TH_ACK) ? "A" : "",
+               (flags & TH_SYN) ? "S" : "",
+               (flags & TH_RST) ? "R" : "",
+               (flags & TH_FIN) ? "F" : "",
+               (flags & TH_URG) ? "U" : "",
+               (flags & TH_PUSH)? "P" : "",
+               (flags & TH_ECE) ? "E" : "",
+               (flags & TH_CWR) ? "C" : "");
+
+    switch (proto) {
+        case IPPROTO_ICMP:
+        case IPPROTO_ICMPV6:
+        case IPPROTO_IGMP:
+            printf(" %u:%u", sport, dport);
+    }
+
+    if (frag)
+        printf(" %s%u@%u:%u",
+               frag_offset?"+":"", frag_id, frag_offset, len);
+
+    if (dump_single)
+        printf(" ");
+    else
+        printf("\n");
+
+    if (quiet < 3)
+        dump_func(data, len);
+
+    if (pd_dump)
+        pcap_dump((u_char*)pd_dump, h, p);
+}
+
+int8_t re_match_func(unsigned char *data, uint32_t len) {
 #if USE_PCRE
-    switch(pcre_exec(pattern, 0, data, (int)len, 0, 0, 0, 0)) {
+    switch(pcre_exec(pattern, 0, data, (int32_t)len, 0, 0, 0, 0)) {
         case PCRE_ERROR_NULL:
         case PCRE_ERROR_BADOPTION:
         case PCRE_ERROR_BADMAGIC:
@@ -836,7 +901,7 @@ int re_match_func(char *data, unsigned len) {
             return 0;
     }
 #else
-    switch (re_search(&pattern, data, (int)len, 0, len, 0)) {
+    switch (re_search(&pattern, data, (int32_t)len, 0, len, 0)) {
         case -2:
             perror("she's dead, jim\n");
             clean_exit(-2);
@@ -855,10 +920,9 @@ int re_match_func(char *data, unsigned len) {
     return 1;
 }
 
-
-int bin_match_func(char *data, unsigned len) {
-    signed stop = len - match_len;
-    signed i = 0;
+int8_t bin_match_func(unsigned char *data, uint32_t len) {
+    int32_t stop = len - match_len;
+    int32_t i    = 0;
 
     if (stop < 0)
         return 0;
@@ -878,20 +942,19 @@ int bin_match_func(char *data, unsigned len) {
 }
 
 
-int blank_match_func(char *data, unsigned len) {
+int8_t blank_match_func(unsigned char *data, uint32_t len) {
     if (max_matches)
         matches++;
 
     return 1;
 }
 
-
-void dump_byline(char *data, unsigned len) {
+void dump_byline(unsigned char *data, uint32_t len) {
     if (len > 0) {
-        const char *s = data;
+        const unsigned char *s = data;
 
         while (s < data + len) {
-            printf("%c", (*s == '\n' || isprint((unsigned char)*s))? (char)*s : nonprint_char);
+            printf("%c", (*s == '\n' || isprint(*s)) ? *s : nonprint_char);
             s++;
         }
 
@@ -899,12 +962,12 @@ void dump_byline(char *data, unsigned len) {
     }
 }
 
-void dump_unwrapped(char *data, unsigned len) {
+void dump_unwrapped(unsigned char *data, uint32_t len) {
     if (len > 0) {
-        const char *s = data;
+        const unsigned char *s = data;
 
         while (s < data + len) {
-            printf("%c", isprint((unsigned char)*s) ? (char)*s : nonprint_char);
+            printf("%c", isprint(*s) ? *s : nonprint_char);
             s++;
         }
 
@@ -912,19 +975,20 @@ void dump_unwrapped(char *data, unsigned len) {
     }
 }
 
-void dump_formatted(char *data, unsigned len) {
+void dump_formatted(unsigned char *data, uint32_t len) {
     if (len > 0) {
-        unsigned width = show_hex?16:(ws_col-5);
-        char *str = data;
-        unsigned j, i = 0;
+        unsigned char *str = data;
+             uint8_t width = show_hex ? 16 : (ws_col-5);
+                uint32_t i = 0,
+                         j = 0;
 
         while (i < len) {
             printf("  ");
 
             if (show_hex)
                 for (j = 0; j < width; j++) {
-                    if (i+j < len)
-                        printf("%02x ", (unsigned char)str[j]);
+                    if (i + j < len)
+                        printf("%02x ", str[j]);
                     else printf("   ");
 
                     if ((j+1) % (width/2) == 0)
@@ -932,12 +996,12 @@ void dump_formatted(char *data, unsigned len) {
                 }
 
             for (j = 0; j < width; j++)
-                if (i+j < len)
-                    printf("%c", isprint((unsigned char)str[j]) ? (char)str[j] : nonprint_char);
+                if (i + j < len)
+                    printf("%c", isprint(str[j]) ? str[j] : nonprint_char);
                 else printf(" ");
 
             str += width;
-            i += j;
+            i   += j;
 
             printf("\n");
         }
@@ -945,27 +1009,24 @@ void dump_formatted(char *data, unsigned len) {
 }
 
 char *get_filter_from_string(char *str) {
-    char *mine;
-    unsigned len;
+    char *mine, *s;
+    uint32_t len;
 
     if (!str || !*str)
         return NULL;
 
-    len = (unsigned)strlen(str);
+    len = (uint32_t)strlen(str);
 
-    {
-        char *s;
-        for (s = str; *s; s++)
-            if (*s == '\r' || *s == '\n')
-                *s = ' ';
-    }
+    for (s = str; *s; s++)
+        if (*s == '\r' || *s == '\n')
+            *s = ' ';
 
-    if (!(mine = (char*)malloc(len + sizeof(IP_ONLY))))
+    if (!(mine = (char*)malloc(len + sizeof(BPF_MAIN_FILTER))))
         return NULL;
 
-    memset(mine, 0, len + sizeof(IP_ONLY));
+    memset(mine, 0, len + sizeof(BPF_MAIN_FILTER));
 
-    sprintf(mine, IP_ONLY, str);
+    sprintf(mine, BPF_MAIN_FILTER, str);
 
     return mine;
 }
@@ -973,20 +1034,20 @@ char *get_filter_from_string(char *str) {
 char *get_filter_from_argv(char **argv) {
     char **arg = argv, *theirs, *mine;
     char *from, *to;
-    unsigned len = 0;
+    uint32_t len = 0;
 
     if (!*arg)
         return NULL;
 
     while (*arg)
-        len += (unsigned)strlen(*arg++) + 1;
+        len += (uint32_t)strlen(*arg++) + 1;
 
     if (!(theirs = (char*)malloc(len + 1)) ||
-        !(mine = (char*)malloc(len + sizeof(IP_ONLY))))
+        !(mine = (char*)malloc(len + sizeof(BPF_MAIN_FILTER))))
         return NULL;
 
     memset(theirs, 0, len + 1);
-    memset(mine, 0, len + sizeof(IP_ONLY));
+    memset(mine, 0, len + sizeof(BPF_MAIN_FILTER));
 
     arg = argv;
     to = theirs;
@@ -996,18 +1057,20 @@ char *get_filter_from_argv(char **argv) {
         *(to-1) = ' ';
     }
 
-    sprintf(mine, IP_ONLY, theirs);
+    sprintf(mine, BPF_MAIN_FILTER, theirs);
 
     free(theirs);
     return mine;
 }
 
 
-int strishex(char *str) {
+uint8_t strishex(char *str) {
     char *s;
+
     if ((s = strchr(str, 'x')))
         s++;
-    else s = str;
+    else
+        s = str;
 
     while (*s)
         if (!isxdigit(*s++))
@@ -1022,11 +1085,11 @@ void print_time_absolute(struct pcap_pkthdr *h) {
 
     printf("%02u/%02u/%02u %02u:%02u:%02u.%06u ",
            t->tm_year+1900, t->tm_mon+1, t->tm_mday, t->tm_hour,
-           t->tm_min, t->tm_sec, h->ts.tv_usec);
+           t->tm_min, t->tm_sec, (uint32_t)h->ts.tv_usec);
 }
 
 void print_time_diff(struct pcap_pkthdr *h) {
-    unsigned secs, usecs;
+    uint32_t secs, usecs;
 
     secs = h->ts.tv_sec - prev_ts.tv_sec;
     if (h->ts.tv_usec >= prev_ts.tv_usec)
@@ -1038,21 +1101,21 @@ void print_time_diff(struct pcap_pkthdr *h) {
 
     printf("+%u.%06u ", secs, usecs);
 
-    prev_ts.tv_sec = h->ts.tv_sec;
+    prev_ts.tv_sec  = h->ts.tv_sec;
     prev_ts.tv_usec = h->ts.tv_usec;
 }
 
 void dump_delay_proc_init(struct pcap_pkthdr *h) {
     dump_delay = &dump_delay_proc;
 
-    prev_delay_ts.tv_sec = h->ts.tv_sec;
+    prev_delay_ts.tv_sec  = h->ts.tv_sec;
     prev_delay_ts.tv_usec = h->ts.tv_usec;
 
     dump_delay(h);
 }
 
 void dump_delay_proc(struct pcap_pkthdr *h) {
-    unsigned secs, usecs;
+    uint32_t secs, usecs;
 
     secs = h->ts.tv_sec - prev_delay_ts.tv_sec;
     if (h->ts.tv_usec >= prev_delay_ts.tv_usec)
@@ -1075,7 +1138,7 @@ void dump_delay_proc(struct pcap_pkthdr *h) {
         FD_ZERO(&delay_fds);
         FD_SET(delay_socket, &delay_fds);
 
-        delay_tv.tv_sec = secs;
+        delay_tv.tv_sec  = secs;
         delay_tv.tv_usec = usecs;
 
         if (select(0, &delay_fds, 0, 0, &delay_tv) == -1)
@@ -1089,12 +1152,12 @@ void dump_delay_proc(struct pcap_pkthdr *h) {
     usleep(usecs);
 #endif
 
-    prev_delay_ts.tv_sec = h->ts.tv_sec;
+    prev_delay_ts.tv_sec  = h->ts.tv_sec;
     prev_delay_ts.tv_usec = h->ts.tv_usec;
 }
 
 #if !defined(_WIN32)
-void update_windowsize(int e) {
+void update_windowsize(int32_t e) {
     if (e == 0 && ws_col_forced)
 
         ws_col = ws_col_forced;
@@ -1146,13 +1209,13 @@ void drop_privs(void) {
 }
 #endif
 
-void usage(int e) {
+void usage(int8_t e) {
     printf("usage: ngrep <-"
 #if defined(_WIN32)
            "L"
 #endif
-           "hXViwqpevxlDtTRM> <-IO pcap_dump> <-n num> <-d dev> <-A num>\n"
-           "             <-s snaplen> <-S limitlen> <-W normal|byline|none> <-c cols>\n"
+           "hNXViwqpevxlDtTRM> <-IO pcap_dump> <-n num> <-d dev> <-A num>\n"
+           "             <-s snaplen> <-S limitlen> <-W normal|byline|single|none> <-c cols>\n"
            "             <-P char> <-F file> <match expression> <bpf filter>\n"
            "   -h  is help/usage\n"
            "   -V  is version information\n"
@@ -1176,10 +1239,11 @@ void usage(int e) {
            "   -A  is dump num packets after a match\n"
            "   -s  is set the bpf caplen\n"
            "   -S  is set the limitlen on matched packets\n"
-           "   -W  is set the dump format (normal, byline, none)\n"
+           "   -W  is set the dump format (normal, byline, single, none)\n"
            "   -c  is force the column width to the specified size\n"
            "   -P  is set the non-printable display char to what is specified\n"
            "   -F  is read the bpf filter from the specified file\n"
+           "   -N  is show sub protocol number\n"
 #if defined(_WIN32)
            "   -d  is use specified device (index) instead of the pcap default\n"
            "   -L  is show the winpcap device list index\n"
@@ -1198,21 +1262,23 @@ void version(void) {
 }
 
 
-void clean_exit(int sig) {
+void clean_exit(int32_t sig) {
     struct pcap_stat s;
-    if (!quiet && sig >= 0) printf("exit\n");
+
+    if (quiet < 1 && sig >= 0)
+        printf("exit\n");
 
 #if USE_PCRE
-    if (pattern) pcre_free(pattern);
+    if (pattern)       pcre_free(pattern);
     if (pattern_extra) pcre_free(pattern_extra);
 #else
     if (pattern.translate) free(pattern.translate);
-    if (pattern.fastmap) free(pattern.fastmap);
+    if (pattern.fastmap)   free(pattern.fastmap);
 #endif
 
-    if (bin_data) free(bin_data);
+    if (bin_data)          free(bin_data);
 
-    if (!quiet && sig >= 0 && !read_file
+    if (quiet < 1 && sig >= 0 && !read_file
      && pd && !pcap_stats(pd, &s))
         printf("%u received, %u dropped\n", s.ps_recv, s.ps_drop);
 
@@ -1229,7 +1295,7 @@ void clean_exit(int sig) {
 }
 
 #if defined(_WIN32)
-int win32_initwinsock(void) {
+int8_t win32_initwinsock(void) {
     WORD wVersionRequested = MAKEWORD(2, 0);
     WSADATA wsaData;
 
@@ -1250,7 +1316,7 @@ int win32_initwinsock(void) {
 }
 
 void win32_listdevices(void) {
-    unsigned i = 0;
+    uint32_t i = 0;
     pcap_if_t *alldevs, *d;
     char errbuf[PCAP_ERRBUF_SIZE];
 
@@ -1273,12 +1339,12 @@ void win32_listdevices(void) {
 }
 
 char *win32_usedevice(const char *index) {
-    unsigned idx = atoi(index), i = 0;
+    int32_t idx = atoi(index), i = 0;
     pcap_if_t *alldevs, *d;
     char errbuf[PCAP_ERRBUF_SIZE];
     char *dev = NULL;
 
-    if (idx == 0) {
+    if (idx <= 0) {
         perror("invalid device index");
         clean_exit(-1);
     }
@@ -1292,7 +1358,7 @@ char *win32_usedevice(const char *index) {
         if (++i == idx)
             dev = _strdup(d->name);
 
-    if (i == 0) {
+    if (i <= 0) {
         perror("no known devices");
         clean_exit(-1);
     }
